; Dark Star, ZX Spectrum.

; These sources are reconstructed from a disassembly using the routine names from a pre-release early version.
; The code they generate is not exactly the same as the original binary, but it's pretty close.

; The game remains the same, aside from being frame-locked to take advantage of emulators that can run the Z80 in turbo
; mode. You can disable that by commenting out the option, a few lines below this. I also fixed a minor bug, you can
; restore that if you like.

; Assembly time options. Comment these out to make them ineffective

;bAnalogDefault          equ true                        ; Default to the joystick

;bFrameLock              equ true                        ; We lock to the video frame

;bDontFixBug             equ true                        ; Leave the (minor) bug alone

;bCompressForTape        equ true                        ; Generate compressed data

; Set the memory debugging up, so the emulator doesn't complain if "Trap on error" is set.

; Shortcut for ParaSys

s                       equ CS_Test                     ; Start here
zoParaSysNotEmulate     = true                          ; Tell Zeus we want to use parasys

TurboRegister           equ $07
MHz35                   equ %00
MHz7                    equ %01
MHz14                   equ %10
MHz28                   equ %01

Turbo                   macro(Mode)
                        nextreg TurboRegister,Mode
  mend

                        zeusemulate "NEXT"

                        mem_smc $0000,1                 ; The graphics routines sometimes read/write location 0

                        mem_ro $0001,$3FFF              ; The stars code sometimes reads the ROM (random positions)

; Note that the way these sources plant data in "ds_data" makes a lot of the data memory read only if we don't have this next bit

                        mem_smc $4000,$C000             ; This just marks all the RAM as Read, write or execute.

                        profile $0000,$10000

                        output_para $5b00,Ye_End-$5B00,CS_Test

; Include the equates

                        include "ds_equ.inc"

                        include "ds_data.inc"

                        ; Include the front-end code

                        include "ds_front.inc"

                        ; Include the game-irq code

                        include "ds_irq.inc"

                        ; Include the user-defined key code

                        include "ds_udk.inc"

                        ; Include the maths code

                        include "ds_maths.inc"

                        ; Include various utilities

                        include "ds_utils.inc"

                        ; Include the parasys stub loader

BootParaBase            equ Ye_End

                        include "ParaBootStub.inc"

                        ; The player's code

                        include "ds_player.inc"

                        ; Include the drawing routines

                        include "ds_draw.inc"

                        ; Include the stars

                        include "ds_stars.inc"

                        ; Include the projectiles (player's weapon)

                        include "ds_proj.inc"

                        ; Include the map, etc

                        include "ds_map.inc"

                        ; Include the object code

                        include "ds_objects.inc"

                        ; Include the game-code

                        include "ds_game.inc"

; These commands tell Zeus where to put the code it generates

; Warn if there are any strange options set

        if enabled bAnalogDefault
          zeusprint "                                               WARNING! Analogue Joystick is the default"
        endif

; This command tells Zeus where to put the code it generates. As a szx file...
; Zeus's internal emulator needs you to generate a szx file somewhere in order to emulate correctly

        if def bDontWriteFiles
                        output_szx "dsunusual.szx",$0000,CS_Test       ; The szx file
          zeusprint "********************************************   WARNING! No files written!"
        else
                        output_szx "ds.szx",$0000,CS_Test              ; The szx file
                        output_z80 "ds.z80",$0000,CS_Test              ; The szx file

          zeusprint "                                               Written full version"
        endif

;                        output_bin OutputDir+"\ds_new.bin",$4000,$C000      ; The binary file

; Messy stuff at the end

        if ZEUSVER < 25
            zeuserror "Please upgrade Zeus - the latest version can be found at www.desdes.com/products/oldfiles"
          if ZEUSVER < 25
            zeuserror "There are problems with the Spectrum emulation in this version of Zeus"
          endif
          if ZEUSVER < 11
            zeuserror "This version of Zeus doesn't understand the DG pseudo-op"
          endif
        endif

; Build instead the compressed data for the loader

        if enabled bCompressForTape and (ZEUSVER >= 26)
          zeuscompress $5B00,Save_Limit,CS_Test,$5B00,1  ; This builds a block of data which you load at $5B00 and then call at $5B00 (the 1 means flash border during decompression)
          output_bin "ds_load.bin",$5B00,$5E00           ; The loader data
          zeusprinthex "You need to save the block starting at $5B00 and call it at $5B00. Highest required byte's address stored at $5B02"
        else
;          zeusprinthex "( You need only save memory below: ", Save_Limit,"during tape saves. Only the stack and score-reporting live above: ", Ye_End, "Call ",CS_Test," to start )"
        endif

; Tell Zeus about routines that have data following the call.

        zeusdatafollows ShowStrFollow,PrintStrFollow


