; Reads the keyboard and sends typed characters, while reporting received ones in hex.
; Quick and dirty print and keyscan code stolen from Dark Star's high-score table routines.
; The cursor isn't perfect, it wraps off the end of lines badly, particularly near
; the bottom of the screen...

;bShowInHex              equ true                        ; If enabled, show received chars in hex

        if enabled bShowInHex
AppFilename             equ "Echo2M_H"                  ; What we're called (for file generation)
        else
AppFilename             equ "Echo2M"                    ; What we're called (for file generation)
        endif

                        zeusemulate "NEXT"              ; This is for the Next

; Print control codes
cDel                    equ $08                         ;
cCret                   equ $0D                         ;
cCls                    equ $0C                         ;
cHome                   equ $0E                         ;
cMid                    equ $0F                         ;
cClr                    equ $1B                         ;

; Print colour select characters

cBlack                  equ $00                         ;
cBlue                   equ $01                         ;
cRed                    equ $02                         ;
cMagenta                equ $03                         ;
cGreen                  equ $04                         ;
cCyan                   equ $05                         ;
cYellow                 equ $06                         ;
cWhite                  equ $07                         ;

                        include "Next_Equ.inc"          ;

TurboRegister           equ $07
MHz35                   equ %00
MHz7                    equ %01
MHz14                   equ %10
MHz28                   equ %01

Turbo                   macro(Mode)
                        nextreg TurboRegister,Mode
  mend

                        org $9000                       ; Start of application
AppEntry                equ *                           ; Replace these lines with your code

Start                   di                              ; Make sure we're not interrupted
                        ld sp,$C000                     ; Set the stack to the top of memory

                        Turbo(MHz14)                    ; Let's go fast...

                        ; Set 2M baud

                        ld bc,$243B                     ; Now adjust for the set Video timing.
                        ld a,17                         ;
                        out (c),a                       ;

                        inc b                           ;
                        in a,(c)                        ;

                        ld hl,BaudTable                 ;
                        add a,l                         ;
                        ld l,a                          ;

                        adc a,h                         ;
                        sub l                           ;
                        ld h,a                          ;

                        ld a,(hl)                       ;

                        ld bc,UART_SetBaud              ;

                        out(c),a                        ;

                        ld a,%1 0000000                 ; Top bit set
                        out(c),a                        ;

                        call InitKbd                    ; Setup the keyboard

                        ld a,cWhite                     ; Colur:=White
                        call Print                      ;
                        ld a,cCls                       ; Clear the screen
                        call Print                      ;

                        call UART_Flush                 ; Clear it

Echo                    call ReadKey                    ; Look for a key

                        call UART_SendChar              ;

                        jr Echo                         ;

; 2M baud table

BaudTable               db 14,14,15,15,16,16,17,14      ; 2000000

                        ; We've timed out

TimeOut                 ld hl,szMessage                 ; Send this

TOStr                   ld a,(hl)                       ;
                        inc hl                          ;
                        or a                            ;
                        jr z Echo                       ;

                        call UART_SendChar              ;

                        jr TOStr                        ; Loop

szMessage               db "Hello!",$0D,$0A,0           ;

; Clear the Rx buffer

UART_Flush              ld bc,UART_GetStatus            ; Read it while there are characters waiting
                        in a,(c)                        ;
                        and UART_mRX_DATA_READY         ;
                        ret z                           ; Nothing waiting, exit

                        ld bc,UART_RxD                  ; Data port (could just load B)
                        in a,(c)                        ; Read it (could perhaps read a few more for speed)
                        jr UART_Flush                   ; Loop

; Wait for a character and return it in A (screws BC,HL)
; Returns NC for time-out

UART_GetChar_TO         ld hl,200                       ; Time-out
                        ld bc,UART_GetStatus            ; Are there any characters waiting?

UART_GC_Wait            in a,(c)                        ;
                        and UART_mRX_DATA_READY         ;
                        jr nz UART_GC1                  ; Yes, read them

                        ; Dec the timer

                        dec hl                          ; Time out?
                        ld a,l                          ;
                        or h                            ;
                        jr nz UART_GC_Wait              ;

                        ret                             ; Time-out

UART_GC1                ld bc,UART_RxD                  ; Data port (could just load B)
                        in a,(c)                        ; Read it (could perhaps read a few more for speed)
                        scf                             ; CF = Have char
                        ret                             ; Done

; Wait for the UART to be available and send A (screws BC)

UART_SendChar           push bc                         ; Save BC
                        push af                         ; Save A

                        ld bc,UART_GetStatus            ; Is the Tx available?
UART_SendCharLp         in a,(c)                        ;
                        and UART_mTX_BUSY               ;
                        jr nz UART_SendCharLp           ; Yes, wait

                        pop af                          ; Restore A
                        out (c),a                       ; Transmit it

                        pop bc                          ; Restore
                        ret                             ; Done

; Read a key...

InitKbd                 LD HL,KeyCounts                 ;
                        LD DE,KeyCounts+1               ;
                        LD BC,nKeys-1                   ;
                        LD (HL),$09                     ;
                        LDIR                            ;
                        LD HL,200                       ;
                        LD (KeyRepSpeed1),HL            ;
                        LD HL,30                        ;
                        LD (KeyRepSpeed2),HL            ;
                        RET                             ;

ReadKey                 PUSH IX                         ;
                        PUSH HL                         ;
                        PUSH DE                         ;
                        PUSH BC                         ;
                        PUSH AF                         ;
                        EX AF,AF'                       ;;
                        PUSH AF                         ;
                        LD HL,KeyCounts                 ;
                        SET 7,(HL)                      ;
                        LD HL,SymbolShift               ;
                        SET 7,(HL)                      ;
                        CALL ReadKeys                   ;
                        DEC A                           ;
                        LD B,A                          ;
                        XOR A                           ;
                        LD HL,KeyCounts                 ;
                        BIT 6,(HL)                      ;
                        JR Z RK_Lp1                     ;

                        ADD A,nKeys                     ;
RK_Lp1                  LD HL,SymbolShift               ;
                        BIT 6,(HL)                      ;
                        JR Z RK_Lp2                     ;

                        ADD A,nKeys*2                   ;
RK_Lp2                  ADD A,B                         ;
                        LD E,A                          ;
                        LD D,$00                        ;
                        LD HL,KTab_0                    ;
                        ADD HL,DE                       ;
                        POP AF                          ;
                        EX AF,AF'                       ;;
                        POP AF                          ;
                        LD A,(HL)                       ;
                        POP BC                          ;
                        POP DE                          ;
                        POP HL                          ;
                        POP IX                          ;
                        RET                             ;

ReadKeys                LD A,$01                        ;
                        LD (CursorSpeed),A              ;
                        LD (CursorType),A               ;

ReadKeyLp               call UART_GetChar_TO            ; Read one
                        jr nc RKTimeOut                 ; Failed

                        push af                         ; Save
                        call CursorOff                  ; Turn it off
                        pop af                          ; Restore

        if enabled bShowInHex                           ;
                        call PrintHex_A                 ; Show it in hex
        else                                            ;
                        call Print                      ; Show it as a character
        endif                                           ;

                        jr ReadKeys                     ; Next

RKTimeOut               CALL CursorInv                  ;
                        CALL ScanKeys                   ;
                        JR NZ RKNewKey                  ;

                        LD A,(CurrKey)                  ;
                        OR A                            ;
                        JR Z ReadKeyLp                  ;

                        LD HL,(CurrKeyAddr)             ;
                        BIT 6,(HL)                      ;
                        JR Z RKCancelOld                ;

                        LD HL,(KeyRepDelay)             ;
                        DEC HL                          ;
                        LD (KeyRepDelay),HL             ;
                        LD A,L                          ;
                        OR H                            ;
                        JR NZ ReadKeyLp                 ;

                        LD HL,(KeyRepSpeed2)            ;
RK1                     LD (KeyRepDelay),HL             ;
                        CALL CursorOff                  ;
                        LD A,(CurrKey)                  ;
                        RET                             ;

RKNewKey                LD (CurrKey),A                  ;
                        LD HL,(KeyRepSpeed1)            ;
                        LD (CurrKeyAddr),IX             ;
                        JR RK1                          ;

RKCancelOld             XOR A                           ;
                        LD (CurrKey),A                  ;
                        JR ReadKeyLp                    ;

CursorOff               LD A,(CursorType)               ;
                        OR A                            ;
                        RET NZ                          ;

                        JP CI_1                         ;

CursorInv               LD A,(CursorSpeed)              ;
                        DEC A                           ;
                        LD (CursorSpeed),A              ;
                        RET NZ                          ;

                        LD A,130                        ;
                        LD (CursorSpeed),A              ;
                        LD A,(CursorType)               ;
                        XOR $01                         ;
                        LD (CursorType),A               ;

CI_1                    LD HL,(pCurY)                   ;
                        LD A,H                          ;
                        ADD A,$07                       ;
                        LD H,A                          ;
                        LD A,(pCurX)                    ;
                        RRCA                            ;
                        RRCA                            ;
                        RRCA                            ;
                        AND $1F                         ;
                        OR L                            ;
                        LD L,A                          ;
                        LD A,(pCurX)                    ;
                        AND $07                         ;
                        LD DE,%1111100000000000         ;
CI_2                    OR A                            ;
                        JR Z CI_3                       ;

                        DEC A                           ;
                        SRL D                           ;
                        RR E                            ;
                        JR CI_2                         ;

CI_3                    LD A,D                          ;
                        XOR (HL)                        ;
                        LD (HL),A                       ;
                        INC L                           ;
                        LD A,E                          ;
                        XOR (HL)                        ;
                        LD (HL),A                       ;
                        RET                             ;

; Scan that poxy keyboard...

ScanKeys                LD BC,$FFFE                     ;
                        LD HL,KeyCounts                 ;
                        LD E,$00                        ;
ScanKeyLp               DEC B                           ;
                        IN A,(C)                        ;
                        RRA                             ;
                        CALL NC ScanKeyPressed          ;
                        CALL C ScanKeyNotPress          ;
                        INC HL                          ;
                        RRA                             ;
                        CALL NC ScanKeyPressed          ;
                        CALL C ScanKeyNotPress          ;
                        INC HL                          ;
                        RRA                             ;
                        CALL NC ScanKeyPressed          ;
                        CALL C ScanKeyNotPress          ;
                        INC HL                          ;
                        RRA                             ;
                        CALL NC ScanKeyPressed          ;
                        CALL C ScanKeyNotPress          ;
                        INC HL                          ;
                        RRA                             ;
                        CALL NC ScanKeyPressed          ;
                        CALL C ScanKeyNotPress          ;
                        INC HL                          ;
                        SCF                             ;
                        RL B                            ;
                        INC B                           ;
                        JR NZ ScanKeyLp                 ;

                        LD A,E                          ;
                        OR A                            ;
                        RET                             ;

ScanKeyPressed          EX AF,AF'                       ;;
                        LD A,(HL)                       ;
                        AND $3F                         ;
                        CP $09                          ;
                        JR Z SKP_1                      ;

                        INC (HL)                        ;
                        EX AF,AF'                       ;;
                        RET                             ;

SKP_1                   BIT 6,(HL)                      ;
                        JR NZ SKP_3                     ;

                        BIT 7,(HL)                      ;
                        JR NZ SKP_2                     ;

                        LD A,E                          ;
                        OR A                            ;
                        JR NZ SKP_3                     ;

                        PUSH HL                         ;
                        POP IX                          ;
                        CALL SKNP_2                     ;
                        LD E,A                          ;
SKP_2                   SET 6,(HL)                      ;
SKP_3                   EX AF,AF'                       ;;
                        RET                             ;

ScanKeyNotPress         EX AF,AF'                       ;;
                        LD A,(HL)                       ;
                        AND $3F                         ;
                        JR Z SKNP_1                     ;

                        DEC (HL)                        ;
                        EX AF,AF'                       ;;
                        RET                             ;

SKNP_1                  RES 6,(HL)                      ;
                        EX AF,AF'                       ;;
                        RET                             ;

SKNP_2                  PUSH DE                         ;
                        LD DE,CursorType                ;
                        OR A                            ;
                        SBC HL,DE                       ;
                        LD A,L                          ;
                        ADD HL,DE                       ;
                        POP DE                          ;
                        RET                             ;

KTab_0                  DEFB $00,$7A,$78,$63,$76,$61,$73,$64;
                        DEFB $66,$67,$71,$77,$65,$72,$74,$31;
                        DEFB $32,$33,$34,$35,$30,$39,$38,$37;
                        DEFB $36,$70,$6F,$69,$75,$79,$0D,$6C;
                        DEFB $6B,$6A,$68,$20,$00,$6D,$6E,$62;

                        DEFB $00,$5A,$58,$43,$56,$41,$53,$44;
                        DEFB $46,$47,$51,$57,$45,$52,$54,$21;
                        DEFB $40,$23,$24,$25,$08,$29,$28,$27;
                        DEFB $26,$50,$4F,$49,$55,$59,$0D,$4C;
                        DEFB $4B,$4A,$48,$20,$00,$4D,$4E,$42;

                        DEFB $00,$3A,$23,$3F,$2F,$7E,$7C,$27;
                        DEFB $7B,$7D,$20,$20,$20,$3C,$3E,$21;
                        DEFB $40,$23,$24,$25,$5F,$29,$28,$27;
                        DEFB $26,$22,$3B,$20,$5D,$5B,$0D,$3D;
                        DEFB $2B,$2D,$5E,$20,$00,$2E,$2C,$2A;

                        DEFB $00,$20,$20,$20,$20,$74,$20,$49;
                        DEFB $20,$67,$61,$20,$67,$69,$72,$47;
                        DEFB $6F,$64,$20,$49,$64,$65,$65,$6E;
                        DEFB $20,$73,$65,$6C,$20,$6C,$64,$61;
                        DEFB $6D,$20,$6F,$20,$00,$20,$20,$20;


; Print the zero-terminated string pointed to by HL

PrintSZ_DE              push de,af                      ;

PSZDE_Lp                ld a,(de)                       ;
                        inc de                          ;
                        or a                            ;
                        jr z,PSZDE_X                    ;

                        call Print                      ;
                        jr PSZDE_Lp                     ;

PSZDE_X                 pop af,de                       ;
                        ret                             ; Done

; Print the string following the call in memory

PrintStrFollow          proc                            ;

                        ex (sp),hl                      ;
                        push af                         ;

Loop                    ld a,(hl)                       ;
                        inc hl                          ;
                        or a                            ;
                        jr z,Exit                       ;

                        call Print                      ;

                        jr Loop                         ;

Exit                    pop af                          ;
                        ex (sp),hl                      ;

                        retp                            ; (RET)


; Print the contents of HL in hex

PrintHex_HL             ld a,h                          ;
                        call PrintHex_A                 ;

                        ld a,l                          ;

                        ; Print the contents of A in hex

PrintHex_A              call PH2                        ;

PH2                     rrca                            ;
                        rrca                            ;
                        rrca                            ;
                        rrca                            ;

                        push af                         ;

                        and $0F                         ;

                        add a,"0"                       ;
                        cp "9"+1                        ;
                        jr c,PH3                        ;

                        add a,"A"-("9"+1)               ;

PH3                     call Print                      ;

                        pop af                          ;
                        ret                             ;

; Print a space

Print_Sp                push af                         ; Save

                        ld a,' '                        ; Print a space
                        call Print                      ;

                        pop af                          ; Done
                        ret                             ;

; A print character routine. Proportionally spaced characters

Print                   push ix,hl,de,bc,af             ; Save them

                        cp cWhite+1                     ;
                        jr c,PrintColour                ;

                        cp cCls                         ;
                        jr z,PrintCls                   ;

                        cp cCret                        ;
                        jr z,PrintCret                  ;

                        cp cHome                        ;
                        jr z,PrintHome                  ;

                        cp cMid                         ;
                        jr z,PrintMid                   ;

                        ld h,PrintWidths/256            ;
                        ld l,a                          ;

                        ld a,(pCurX)                    ;
                        add a,(hl)                      ;
                        call c,PrintDoCret              ;

                        call PrintOutChar               ;

                        ld a,(pCurX)                    ;
                        add a,(hl)                      ;
                        ld (pCurX),a                    ;

PrintExit               pop af,bc,de,hl,ix              ;
                        ret                             ;

; Set a colour

PrintColour             cp 4                            ; Above green don't make them bright
                        jr nc,PrintCol1                 ;

                        or $40                          ; Bright

PrintCol1               ld (CurCol),a                   ;
                        jr PrintExit                    ;

; Carriage return

PrintCret               call PrintDoCret                ;
                        jr PrintExit                    ;

; Clear screen - this clears it slowly with a fade effect

PrintCls                ld e,$00                        ;
                        ld hl,$5800                     ;
                        ld bc,$02FF                     ;

pC1                     ld a,(hl)                       ;
                        and $07                         ;
                        jr z,pC2                        ;

                        set 0,e                         ;
                        dec (hl)                        ;
pC2                     cpi                             ;
                        jp pe,pC1                       ;

                        call Delay                      ;

                        bit 0,e                         ;
                        jr nz,PrintCls                  ;

                        ld hl,$4000                     ;
                        ld de,$4001                     ;
                        ld bc,$1800                     ;

                        xor a                           ;
                        ld (hl),a                       ;

                        ldir                            ;

                        ld a,(CurCol)                   ;
                        ld (hl),a                       ;

                        ld bc,$02FF                     ;

                        ldir                            ;

                        xor a                           ;
                        out ($FE),a                     ;

PrintHome               xor a                           ;
                        ld (pCurX),a                    ;
                        ld (ppCurY),a                   ;

                        ld a,cWhite                     ;
                        ld (CurCol),a                   ;

                        ld hl,$4000                     ;
                        ld (pCurY),hl                   ;

                        jr PrintExit                    ;

PrintMid                ld a,160                        ;
                        ld (pCurX),a                    ;

                        jr PrintExit                    ;

; A suitable delay for the fade

Delay                   push bc                         ;
                        ld bc,1000                      ;
Del1                    cpi                             ;
                        dec hl                          ;
                        jp pe,Del1                      ;
                        pop bc                          ;
                        ret                             ;

; Do the carriage return

PrintDoCret             ld a,(ppCurY)                   ;
                        inc a                           ;

PrintSetUpY             cp 24                           ; Gone off the bottom?
                        jr c PSUY_Ok                    ; No, it's Okay

                        ; Scroll the screen up one line

                        call ScrollUp                   ;

                        ld a,23                         ; Bottom line

PSUY_Ok                 ld (ppCurY),a                   ;

                        and $1F                         ;

                        push hl                         ;

                        ld l,a                          ;
                        ld h,$00                        ;

                        add hl,hl                       ;

                        ld de,YTable                    ;
                        add hl,de                       ;

                        ld a,(hl)                       ;
                        inc hl                          ;
                        ld h,(hl)                       ;
                        ld l,a                          ;

                        ld (pCurY),hl                   ;

                        xor a                           ;
                        ld (pCurX),a                    ;

                        pop hl                          ;
                        ret                             ;

; Output the character

PrintOutChar            bit 7,l                         ; characters $00..$7F only
                        ret nz                          ;

                        push hl                         ;

                        ld h,0                          ;

                        add hl,hl                       ;
                        add hl,hl                       ;
                        add hl,hl                       ;

                        ld de,PrintChars-256            ;
                        add hl,de                       ;

                        push hl                         ;
                        pop ix                          ;

                        ld a,(pCurX)                    ;

                        rrca                            ;
                        rrca                            ;
                        rrca                            ;
                        and $1F                         ;

                        ld de,(pCurY)                   ;

                        or e                            ;
                        ld e,a                          ;

                        push de                         ;

                        ld a,d                          ;
                        rrca                            ;
                        rrca                            ;
                        rrca                            ;
                        and $03                         ;
                        or $58                          ;
                        ld d,a                          ;

                        ld a,(CurCol)                   ;
                        ld (de),a                       ;
                        inc de                          ;
                        ld (de),a                       ;

                        pop de                          ;

                        ld a,(pCurX)                    ;
                        and $07                         ;
                        ld c,a                          ;

                        ld b,$08                        ;

OutCharL                ld h,(ix)                       ;
                        ld l,$00                        ;

                        ld a,c                          ;
                        or a                            ;
                        jr z,OCL1                       ;

OCL2                    srl h                           ;
                        rr l                            ;
                        dec a                           ;
                        jr nz,OCL2                      ;

OCL1                    ld a,(de)                       ;
                        xor h                           ;
                        ld (de),a                       ;

                        inc e                           ;

                        ld a,(de)                       ;
                        xor l                           ;
                        ld (de),a                       ;

                        dec e                           ;

                        inc ix                          ;
                        inc d                           ;

                        djnz OutCharL                   ;

                        pop hl                          ;
                        ret                             ;

ScrollUp                push ix,hl,de,bc                ; Save

                        ld ix,YTable                    ;
SU_Lp                   ld hl,(ix+2)                    ;

                        ld a,l                          ;
                        or h                            ;
                        jr z SU_Moved                   ;

                        ld de,(ix)                      ;

                        inc ix                          ;
                        inc ix                          ;

                        push hl,de                      ;

                        ld a,d                          ;
                        rrca                            ;
                        rrca                            ;
                        rrca                            ;
                        and $03                         ;
                        or $58                          ;
                        ld d,a                          ;

                        ld a,h                          ;
                        rrca                            ;
                        rrca                            ;
                        rrca                            ;
                        and $03                         ;
                        or $58                          ;
                        ld h,a                          ;

                        ld bc,32                        ;
                        ldir                            ;
                        pop de,hl                       ;



SU_Lp2                  push hl,de                      ;
                        ld bc,32                        ;
                        ldir                            ;
                        pop de,hl                       ;

                        inc h                           ;
                        inc d                           ;

                        ld a,h                          ;
                        and %00000111                   ;
                        jr nz SU_Lp2                    ;

                        jr SU_Lp                        ; Loop

                        ; Clear the bottom line

SU_Moved                ld hl,$50E0                     ; Clear the bottom line
                        call ClearLine                  ;

                        pop bc,de,hl,ix                 ;
                        ret                             ;

                        ; Clear a line

ClearLine               push hl                         ;

                        ld d,h                          ;
                        ld e,l                          ;
                        inc de                          ;

                        ld (hl),0                       ;
                        ld bc,31                        ;
                        ldir                            ;

                        pop hl                          ;

                        inc h                           ;

                        ld a,h                          ;
                        and %00000111                   ;
                        jr nz ClearLine                 ;

                        ret                             ; Done

; Screen addresses for the character rows

YTable                  defw $4000,$4020,$4040,$4060    ;
                        defw $4080,$40A0,$40C0,$40E0    ;
                        defw $4800,$4820,$4840,$4860    ;
                        defw $4880,$48A0,$48C0,$48E0    ;
                        defw $5000,$5020,$5040,$5060    ;
                        defw $5080,$50A0,$50C0,$50E0    ;
                        defw 0,0,0,0,0,0,0,0            ; So running off the end isn't fatal

; Page bounded data

                        align 256                       ; Put these on page boundaries

; The CRC data rearranged for faster access

crcpages                DEFB $00,$77,$EE,$99,$07,$70,$E9,$9E;
                        DEFB $0E,$79,$E0,$97,$09,$7E,$E7,$90;
                        DEFB $1D,$6A,$F3,$84,$1A,$6D,$F4,$83;
                        DEFB $13,$64,$FD,$8A,$14,$63,$FA,$8D;
                        DEFB $3B,$4C,$D5,$A2,$3C,$4B,$D2,$A5;
                        DEFB $35,$42,$DB,$AC,$32,$45,$DC,$AB;
                        DEFB $26,$51,$C8,$BF,$21,$56,$CF,$B8;
                        DEFB $28,$5F,$C6,$B1,$2F,$58,$C1,$B6;
                        DEFB $76,$01,$98,$EF,$71,$06,$9F,$E8;
                        DEFB $78,$0F,$96,$E1,$7F,$08,$91,$E6;
                        DEFB $6B,$1C,$85,$F2,$6C,$1B,$82,$F5;
                        DEFB $65,$12,$8B,$FC,$62,$15,$8C,$FB;
                        DEFB $4D,$3A,$A3,$D4,$4A,$3D,$A4,$D3;
                        DEFB $43,$34,$AD,$DA,$44,$33,$AA,$DD;
                        DEFB $50,$27,$BE,$C9,$57,$20,$B9,$CE;
                        DEFB $5E,$29,$B0,$C7,$59,$2E,$B7,$C0;
                        DEFB $ED,$9A,$03,$74,$EA,$9D,$04,$73;
                        DEFB $E3,$94,$0D,$7A,$E4,$93,$0A,$7D;
                        DEFB $F0,$87,$1E,$69,$F7,$80,$19,$6E;
                        DEFB $FE,$89,$10,$67,$F9,$8E,$17,$60;
                        DEFB $D6,$A1,$38,$4F,$D1,$A6,$3F,$48;
                        DEFB $D8,$AF,$36,$41,$DF,$A8,$31,$46;
                        DEFB $CB,$BC,$25,$52,$CC,$BB,$22,$55;
                        DEFB $C5,$B2,$2B,$5C,$C2,$B5,$2C,$5B;
                        DEFB $9B,$EC,$75,$02,$9C,$EB,$72,$05;
                        DEFB $95,$E2,$7B,$0C,$92,$E5,$7C,$0B;
                        DEFB $86,$F1,$68,$1F,$81,$F6,$6F,$18;
                        DEFB $88,$FF,$66,$11,$8F,$F8,$61,$16;
                        DEFB $A0,$D7,$4E,$39,$A7,$D0,$49,$3E;
                        DEFB $AE,$D9,$40,$37,$A9,$DE,$47,$30;
                        DEFB $BD,$CA,$53,$24,$BA,$CD,$54,$23;
                        DEFB $B3,$C4,$5D,$2A,$B4,$C3,$5A,$2D;
                        DEFB $00,$07,$0E,$09,$6D,$6A,$63,$64;
                        DEFB $DB,$DC,$D5,$D2,$B6,$B1,$B8,$BF;
                        DEFB $B7,$B0,$B9,$BE,$DA,$DD,$D4,$D3;
                        DEFB $6C,$6B,$62,$65,$01,$06,$0F,$08;
                        DEFB $6E,$69,$60,$67,$03,$04,$0D,$0A;
                        DEFB $B5,$B2,$BB,$BC,$D8,$DF,$D6,$D1;
                        DEFB $D9,$DE,$D7,$D0,$B4,$B3,$BA,$BD;
                        DEFB $02,$05,$0C,$0B,$6F,$68,$61,$66;
                        DEFB $DC,$DB,$D2,$D5,$B1,$B6,$BF,$B8;
                        DEFB $07,$00,$09,$0E,$6A,$6D,$64,$63;
                        DEFB $6B,$6C,$65,$62,$06,$01,$08,$0F;
                        DEFB $B0,$B7,$BE,$B9,$DD,$DA,$D3,$D4;
                        DEFB $B2,$B5,$BC,$BB,$DF,$D8,$D1,$D6;
                        DEFB $69,$6E,$67,$60,$04,$03,$0A,$0D;
                        DEFB $05,$02,$0B,$0C,$68,$6F,$66,$61;
                        DEFB $DE,$D9,$D0,$D7,$B3,$B4,$BD,$BA;
                        DEFB $B8,$BF,$B6,$B1,$D5,$D2,$DB,$DC;
                        DEFB $63,$64,$6D,$6A,$0E,$09,$00,$07;
                        DEFB $0F,$08,$01,$06,$62,$65,$6C,$6B;
                        DEFB $D4,$D3,$DA,$DD,$B9,$BE,$B7,$B0;
                        DEFB $D6,$D1,$D8,$DF,$BB,$BC,$B5,$B2;
                        DEFB $0D,$0A,$03,$04,$60,$67,$6E,$69;
                        DEFB $61,$66,$6F,$68,$0C,$0B,$02,$05;
                        DEFB $BA,$BD,$B4,$B3,$D7,$D0,$D9,$DE;
                        DEFB $64,$63,$6A,$6D,$09,$0E,$07,$00;
                        DEFB $BF,$B8,$B1,$B6,$D2,$D5,$DC,$DB;
                        DEFB $D3,$D4,$DD,$DA,$BE,$B9,$B0,$B7;
                        DEFB $08,$0F,$06,$01,$65,$62,$6B,$6C;
                        DEFB $0A,$0D,$04,$03,$67,$60,$69,$6E;
                        DEFB $D1,$D6,$DF,$D8,$BC,$BB,$B2,$B5;
                        DEFB $BD,$BA,$B3,$B4,$D0,$D7,$DE,$D9;
                        DEFB $66,$61,$68,$6F,$0B,$0C,$05,$02;
                        DEFB $00,$30,$61,$51,$C4,$F4,$A5,$95;
                        DEFB $88,$B8,$E9,$D9,$4C,$7C,$2D,$1D;
                        DEFB $10,$20,$71,$41,$D4,$E4,$B5,$85;
                        DEFB $98,$A8,$F9,$C9,$5C,$6C,$3D,$0D;
                        DEFB $20,$10,$41,$71,$E4,$D4,$85,$B5;
                        DEFB $A8,$98,$C9,$F9,$6C,$5C,$0D,$3D;
                        DEFB $30,$00,$51,$61,$F4,$C4,$95,$A5;
                        DEFB $B8,$88,$D9,$E9,$7C,$4C,$1D,$2D;
                        DEFB $41,$71,$20,$10,$85,$B5,$E4,$D4;
                        DEFB $C9,$F9,$A8,$98,$0D,$3D,$6C,$5C;
                        DEFB $51,$61,$30,$00,$95,$A5,$F4,$C4;
                        DEFB $D9,$E9,$B8,$88,$1D,$2D,$7C,$4C;
                        DEFB $61,$51,$00,$30,$A5,$95,$C4,$F4;
                        DEFB $E9,$D9,$88,$B8,$2D,$1D,$4C,$7C;
                        DEFB $71,$41,$10,$20,$B5,$85,$D4,$E4;
                        DEFB $F9,$C9,$98,$A8,$3D,$0D,$5C,$6C;
                        DEFB $83,$B3,$E2,$D2,$47,$77,$26,$16;
                        DEFB $0B,$3B,$6A,$5A,$CF,$FF,$AE,$9E;
                        DEFB $93,$A3,$F2,$C2,$57,$67,$36,$06;
                        DEFB $1B,$2B,$7A,$4A,$DF,$EF,$BE,$8E;
                        DEFB $A3,$93,$C2,$F2,$67,$57,$06,$36;
                        DEFB $2B,$1B,$4A,$7A,$EF,$DF,$8E,$BE;
                        DEFB $B3,$83,$D2,$E2,$77,$47,$16,$26;
                        DEFB $3B,$0B,$5A,$6A,$FF,$CF,$9E,$AE;
                        DEFB $C2,$F2,$A3,$93,$06,$36,$67,$57;
                        DEFB $4A,$7A,$2B,$1B,$8E,$BE,$EF,$DF;
                        DEFB $D2,$E2,$B3,$83,$16,$26,$77,$47;
                        DEFB $5A,$6A,$3B,$0B,$9E,$AE,$FF,$CF;
                        DEFB $E2,$D2,$83,$B3,$26,$16,$47,$77;
                        DEFB $6A,$5A,$0B,$3B,$AE,$9E,$CF,$FF;
                        DEFB $F2,$C2,$93,$A3,$36,$06,$57,$67;
                        DEFB $7A,$4A,$1B,$2B,$BE,$8E,$DF,$EF;
                        DEFB $00,$96,$2C,$BA,$19,$8F,$35,$A3;
                        DEFB $32,$A4,$1E,$88,$2B,$BD,$07,$91;
                        DEFB $64,$F2,$48,$DE,$7D,$EB,$51,$C7;
                        DEFB $56,$C0,$7A,$EC,$4F,$D9,$63,$F5;
                        DEFB $C8,$5E,$E4,$72,$D1,$47,$FD,$6B;
                        DEFB $FA,$6C,$D6,$40,$E3,$75,$CF,$59;
                        DEFB $AC,$3A,$80,$16,$B5,$23,$99,$0F;
                        DEFB $9E,$08,$B2,$24,$87,$11,$AB,$3D;
                        DEFB $90,$06,$BC,$2A,$89,$1F,$A5,$33;
                        DEFB $A2,$34,$8E,$18,$BB,$2D,$97,$01;
                        DEFB $F4,$62,$D8,$4E,$ED,$7B,$C1,$57;
                        DEFB $C6,$50,$EA,$7C,$DF,$49,$F3,$65;
                        DEFB $58,$CE,$74,$E2,$41,$D7,$6D,$FB;
                        DEFB $6A,$FC,$46,$D0,$73,$E5,$5F,$C9;
                        DEFB $3C,$AA,$10,$86,$25,$B3,$09,$9F;
                        DEFB $0E,$98,$22,$B4,$17,$81,$3B,$AD;
                        DEFB $20,$B6,$0C,$9A,$39,$AF,$15,$83;
                        DEFB $12,$84,$3E,$A8,$0B,$9D,$27,$B1;
                        DEFB $44,$D2,$68,$FE,$5D,$CB,$71,$E7;
                        DEFB $76,$E0,$5A,$CC,$6F,$F9,$43,$D5;
                        DEFB $E8,$7E,$C4,$52,$F1,$67,$DD,$4B;
                        DEFB $DA,$4C,$F6,$60,$C3,$55,$EF,$79;
                        DEFB $8C,$1A,$A0,$36,$95,$03,$B9,$2F;
                        DEFB $BE,$28,$92,$04,$A7,$31,$8B,$1D;
                        DEFB $B0,$26,$9C,$0A,$A9,$3F,$85,$13;
                        DEFB $82,$14,$AE,$38,$9B,$0D,$B7,$21;
                        DEFB $D4,$42,$F8,$6E,$CD,$5B,$E1,$77;
                        DEFB $E6,$70,$CA,$5C,$FF,$69,$D3,$45;
                        DEFB $78,$EE,$54,$C2,$61,$F7,$4D,$DB;
                        DEFB $4A,$DC,$66,$F0,$53,$C5,$7F,$E9;
                        DEFB $1C,$8A,$30,$A6,$05,$93,$29,$BF;
                        DEFB $2E,$B8,$02,$94,$37,$A1,$1B,$8D;

; The proportional font (from Invasion Of The Body Snatchers)

PrintChars              defb $00,$00,$00,$00,$00,$00,$00,$00;
                        defb $80,$80,$80,$80,$80,$00,$80,$00;
                        defb $A0,$A0,$A0,$00,$00,$00,$00,$00;
                        defb $50,$50,$F8,$50,$F8,$50,$50,$00;
                        defb $20,$78,$A0,$70,$28,$F0,$20,$00;
                        defb $C0,$C8,$10,$20,$40,$98,$18,$00;
                        defb $40,$A0,$A0,$40,$A8,$90,$68,$00;
                        defb $80,$80,$80,$00,$00,$00,$00,$00;
                        defb $20,$40,$80,$80,$80,$40,$20,$00;
                        defb $80,$40,$20,$20,$20,$40,$80,$00;
                        defb $20,$A8,$70,$20,$70,$A8,$20,$00;
                        defb $00,$20,$20,$F8,$20,$20,$00,$00;
                        defb $00,$00,$00,$00,$40,$40,$80,$00;
                        defb $00,$00,$00,$F8,$00,$00,$00,$00;
                        defb $00,$00,$00,$00,$00,$00,$80,$00;
                        defb $00,$08,$10,$20,$40,$80,$00,$00;
                        defb $70,$88,$98,$A8,$C8,$88,$70,$00;
                        defb $20,$60,$20,$20,$20,$20,$70,$00;
                        defb $70,$88,$08,$30,$40,$80,$F8,$00;
                        defb $F8,$08,$10,$30,$08,$88,$70,$00;
                        defb $10,$30,$50,$90,$F8,$10,$10,$00;
                        defb $F8,$80,$F0,$08,$08,$88,$70,$00;
                        defb $38,$40,$80,$F0,$88,$88,$70,$00;
                        defb $F8,$08,$10,$20,$40,$40,$40,$00;
                        defb $70,$88,$88,$70,$88,$88,$70,$00;
                        defb $70,$88,$88,$78,$08,$10,$E0,$00;
                        defb $00,$00,$80,$00,$80,$00,$00,$00;
                        defb $00,$00,$40,$00,$40,$40,$80,$00;
                        defb $10,$20,$40,$80,$40,$20,$10,$00;
                        defb $00,$00,$F8,$00,$F8,$00,$00,$00;
                        defb $80,$40,$20,$10,$20,$40,$80,$00;
                        defb $70,$88,$10,$20,$20,$00,$20,$00;
                        defb $70,$88,$A8,$B8,$B0,$80,$78,$00;
                        defb $20,$50,$88,$88,$F8,$88,$88,$00;
                        defb $F0,$88,$88,$F0,$88,$88,$F0,$00;
                        defb $70,$88,$80,$80,$80,$88,$70,$00;
                        defb $F0,$88,$88,$88,$88,$88,$F0,$00;
                        defb $F8,$80,$80,$F0,$80,$80,$F8,$00;
                        defb $F8,$80,$80,$F0,$80,$80,$80,$00;
                        defb $78,$80,$80,$80,$98,$88,$78,$00;
                        defb $88,$88,$88,$F8,$88,$88,$88,$00;
                        defb $E0,$40,$40,$40,$40,$40,$E0,$00;
                        defb $08,$08,$08,$08,$08,$88,$70,$00;
                        defb $88,$90,$A0,$C0,$A0,$90,$88,$00;
                        defb $80,$80,$80,$80,$80,$80,$F8,$00;
                        defb $88,$D8,$A8,$A8,$88,$88,$88,$00;
                        defb $88,$88,$C8,$A8,$98,$88,$88,$00;
                        defb $70,$88,$88,$88,$88,$88,$70,$00;
                        defb $F0,$88,$88,$F0,$80,$80,$80,$00;
                        defb $70,$88,$88,$88,$A8,$90,$68,$00;
                        defb $F0,$88,$88,$F0,$A0,$90,$88,$00;
                        defb $70,$88,$80,$70,$08,$88,$70,$00;
                        defb $F8,$20,$20,$20,$20,$20,$20,$00;
                        defb $88,$88,$88,$88,$88,$88,$70,$00;
                        defb $88,$88,$88,$88,$88,$50,$20,$00;
                        defb $88,$88,$88,$A8,$A8,$D8,$88,$00;
                        defb $88,$88,$50,$20,$50,$88,$88,$00;
                        defb $88,$88,$50,$20,$20,$20,$20,$00;
                        defb $F8,$08,$10,$20,$40,$80,$F8,$00;
                        defb $F8,$C0,$C0,$C0,$C0,$C0,$F8,$00;
                        defb $00,$80,$40,$20,$10,$08,$00,$00;
                        defb $F8,$18,$18,$18,$18,$18,$F8,$00;
                        defb $20,$70,$A8,$20,$20,$20,$00,$00;
                        defb $00,$00,$00,$00,$00,$00,$F8,$00;
                        defb $00,$00,$00,$00,$00,$00,$00,$00;
                        defb $00,$00,$68,$98,$88,$98,$68,$00;
                        defb $80,$80,$B0,$C8,$88,$C8,$B0,$00;
                        defb $00,$00,$78,$80,$80,$80,$78,$00;
                        defb $08,$08,$68,$98,$88,$98,$68,$00;
                        defb $00,$00,$70,$88,$F8,$80,$70,$00;
                        defb $20,$40,$40,$E0,$40,$40,$40,$00;
                        defb $00,$00,$78,$88,$98,$68,$08,$70;
                        defb $80,$80,$F0,$88,$88,$88,$88,$00;
                        defb $80,$00,$80,$80,$80,$80,$80,$00;
                        defb $00,$20,$00,$20,$20,$20,$20,$C0;
                        defb $80,$80,$90,$A0,$E0,$90,$88,$00;
                        defb $C0,$40,$40,$40,$40,$40,$E0,$00;
                        defb $00,$00,$D0,$A8,$A8,$A8,$A8,$00;
                        defb $00,$00,$F0,$88,$88,$88,$88,$00;
                        defb $00,$00,$70,$88,$88,$88,$70,$00;
                        defb $00,$00,$F0,$88,$88,$F0,$80,$80;
                        defb $00,$00,$78,$88,$88,$78,$08,$08;
                        defb $00,$00,$B0,$C0,$80,$80,$80,$00;
                        defb $00,$00,$78,$80,$70,$08,$F0,$00;
                        defb $00,$20,$F8,$20,$20,$20,$18,$00;
                        defb $00,$00,$88,$88,$88,$88,$78,$00;
                        defb $00,$00,$88,$88,$50,$50,$20,$00;
                        defb $00,$00,$88,$88,$88,$A8,$50,$00;
                        defb $00,$00,$88,$50,$20,$50,$88,$00;
                        defb $00,$00,$90,$90,$90,$F0,$10,$E0;
                        defb $00,$00,$F8,$10,$20,$40,$F8,$00;
                        defb $30,$40,$40,$80,$40,$40,$30,$00;
                        defb $00,$80,$80,$00,$00,$80,$80,$00;
                        defb $C0,$20,$20,$10,$20,$20,$C0,$00;
                        defb $50,$00,$70,$88,$F8,$80,$70,$00;
                        defb $00,$00,$08,$70,$80,$00,$00,$00;

; Character widths

PrintWidths             defb $07,$09,$09,$09,$09,$09,$09,$03;
                        defb $05,$05,$05,$05,$03,$05,$05,$05;
                        defb $05,$03,$05,$05,$05,$05,$09,$09 ; " " ..
                        defb $07,$09,$07,$02,$05,$09,$07,$09;
                        defb $06,$02,$04,$06,$06,$06,$06,$02;
                        defb $04,$04,$06,$06,$03,$06,$02,$06;
                        defb $06,$06,$06,$06,$06,$06,$06,$06;
                        defb $06,$06,$02,$03,$05,$06,$05,$06;
                        defb $06,$06,$06,$06,$06,$06,$06,$06;
                        defb $06,$04,$06,$06,$06,$06,$06,$06;
                        defb $06,$06,$06,$06,$06,$06,$06,$06;
                        defb $06,$06,$06,$06,$06,$06,$06,$06;
                        defb $06,$06,$06,$06,$06,$06,$04,$06;
                        defb $06,$02,$04,$06,$04,$06,$06,$06;
                        defb $06,$06,$05,$06,$06,$06,$06,$06;
                        defb $06,$05,$06,$05,$02,$05,$06,$06;
        if false                                        ;
                        defb $03,$06,$0C,$12,$18,$1E,$24,$2A ; Wide spaces $80..?
                        defb $30,$36,$3C,$42,$48,$06,$06,$06;
                        defb $01,$02,$03,$04,$05,$06,$06,$06;
                        defb $06,$06,$06,$06,$06,$06,$06,$06;
                        defb $06,$06,$06,$06,$06,$06,$06,$06;
                        defb $06,$06,$06,$06,$06,$06,$06,$06;
                        defb $06,$06,$06,$06,$06,$06,$06,$06;
                        defb $06,$06,$06,$06,$06,$06,$06,$06;
                        defb $06,$06,$06,$06,$06,$06,$06,$06;
                        defb $06,$06,$06,$06,$06,$06,$06,$06;
                        defb $06,$06,$06,$06,$06,$06,$06,$06;
                        defb $06,$06,$06,$06,$06,$06,$06,$06;
                        defb $06,$06,$06,$06,$06,$06,$06,$06;
                        defb $06,$06,$06,$06,$06,$06,$06,$06;
                        defb $06,$06,$06,$06,$06,$06,$06,$06;
                        defb $06,$06,$06,$06,$06,$06,$06,$06;
        endif                                           ;

; Variables

pCurX                   db 0                            ; A logical X position
ppCurY                  db 0                            ; A logical Y position
pCurY                   dw 0                            ; The actual address
CurCol                  db 0                            ; The cursor colour

CurrKey                 DEFS 1                          ;
CurrKeyAddr             DEFW KeyCounts                  ;
KeyRepDelay             DEFW 200                        ;
KeyRepSpeed1            DEFW 200                        ;
KeyRepSpeed2            DEFW 30                         ;
CursorSpeed             DEFB 90                         ;
CursorType              DEFB 0                          ;

nKeys                   equ 40                          ;
KeyCounts               DEFS nKeys-4,$09                ;
SymbolShift             DEFS $4,$09                     ;

; Stop planting code after this. (When generating a tape file we save bytes below here)
AppLast                 equ *-1                         ; The last used byte's address

; Generate some useful debugging commands

                        profile AppEntry,AppLast-AppEntry+1 ; Enable profiling for all the code

; Setup the emulation registers, so Zeus can emulate this code correctly

Zeus_PC                 equ AppEntry                    ; Tell the emulator where to start
Zeus_SP                 equ $FF46                       ; Tell the emulator where to put the stack

; These generate some output files

                        ; Generate a Z80 file, these understand 128K files
                        output_z80 AppFilename+".z80",$0000,AppEntry ; A z80 file

                        output_para AppEntry,AppLast-AppEntry,AppEntry;




